var env    = process.env.NODE_ENV || "development",
  DBconfig = require('../db.config.json')[env],
  config   = require('../deployment.environments.json')[env],
  StatsD   = require('node-dogstatsd').StatsD,
  http     = require('http'),
  https    = require('https'),
  maxSockets;
  
//this is the maximum number of concurrent requests to couchDB
maxSockets = config.maxSockets || 100;
http.globalAgent.maxSockets = https.globalAgent.maxSockets = maxSockets;

//local vars
var winston = require('winston'),
  express   = require('express'),
  moment    = require('moment'),
  monitor   = require('./library/monitor'),
  app       = express();
  
if (!config)   return winston.info('Invalid environment: ' + env);
if (!DBconfig) return winston.info('Invalid DB config: '+env);

datadog = new StatsD(config.datadogURL, config.datadogPort);
db      = require('./library/couchClient')({
  url : DBconfig.protocol+
    '://' + DBconfig.username + 
    ':'   + DBconfig.password + 
    '@'   + DBconfig.host + 
    ':'   + DBconfig.port + 
    '/'   + DBconfig.database,
  //log : function (id, args) {
  //  console.log(id, args);
  //},
  request_defaults : {timeout :30 * 1000}, //30 seconds max for couchDB 
});
gatewayNameToAddress = require('./utils').gatewayNameToAddress;

//set up global debug and cache variables
DEBUG = (process.argv.indexOf('debug')  !== -1) ? true : false;
CACHE = config.redis && config.redis.enabled    ? true : false;

if (process.argv.indexOf('debug')    !== -1) DEBUG = true;
if (process.argv.indexOf('no-cache') !== -1) CACHE = false; 

gatewayList = require('./gateways.json');
  // TODO find permanent location for gateways list
  // should the gateways json file live in couchdb?

DATEARRAY  = ['YYYY', '-MM', '-DD', 'THH', ':mm', ':ssZZ'];
DATEFORMAT = DATEARRAY.join('');
  
var apiRoutes = {
  'offers'                  : require("./routes/offers"),
  'offersexercised'         : require("./routes/offersExercised"),
  'topmarkets'              : require("./routes/topMarkets"),
  'markettraders'           : require("./routes/marketTraders"),
  'accountscreated'         : require("./routes/accountsCreated"),
  'issuercapitalization'    : require("./routes/issuerCapitalization"),
  'currencybalances'        : require("./routes/currencyBalances"),
  'totalnetworkvalue'       : require("./routes/totalNetworkValue"),
  'exchangerates'           : require("./routes/exchangeRates"),
  'valuesent'               : require("./routes/valueSent"),
  'totalvaluesent'          : require("./routes/totalValueSent"),
  'accounttransactions'     : require("./routes/accountTransactions"),
  'accounttransactionstats' : require("./routes/accountTransactionStats"),
  'accountoffersexercised'  : require("./routes/accountOffersExercised"),
  'accounttrust'            : require("./routes/accountTrust"),
  'transactionstats'        : require("./routes/transactionStats"),
  'ledgersclosed'           : require("./routes/ledgersClosed"),
  'historicalmetrics'       : require("./routes/historicalMetrics")
};


// enable CORS
var allowCrossDomain = function(req, res, next) {
  res.header('Access-Control-Allow-Origin', '*');
  res.header('Access-Control-Allow-Methods', 'GET,PUT,POST,DELETE,OPTIONS');
  res.header('Access-Control-Allow-Headers', 'Content-Type, Authorization, Content-Length, X-Requested-With');

  // intercept OPTIONS method
  if ('OPTIONS' == req.method) res.send(200);
  else next();
};

app.use(allowCrossDomain);
app.use(express.json());
app.use(express.urlencoded());
app.post('/api/*', requestHandler);
app.listen(config.port);
winston.info('Listening on port ' + config.port);

// Returns BTC/XPS trades from any issuer.
// Swollow errors and report an empty trade set.
var bitcoinPaysharesTrades = require('./routes/bitcoinPaysharesTrades');
app.get('/api/BTC/XPS/trades', function(req, res) {
  bitcoinPaysharesTrades(req.query, function(err, response) {
    if(err) {
      winston.error(err, " - /api/BTC/XPS/trades", "(Spoof empty trade set) 200");
      res.send(200, []);
      return;
    }

    res.send(200, response);
  });
});

// Set required ETag for orderbook requests to a constant value.
app.head('/api/BTC/XPS/orderbook', function(req, res) {
  res.set('ETag', 'not-supported');
  res.send(200);
});

// Returns an empty order book, since orderbooks require an issuer.
app.get('/api/BTC/XPS/orderbook', function(req, res) {
  res.set('ETag', 'not-supported');
  res.send(200, {bids: [], asks: []});
});

// Returns BTC/USD trades from any issuer.
// Swollow errors and report an empty trade set.
var bitcoinUSDollarTrades = require('./routes/bitcoinUSDollarTrades');
app.get('/api/BTC/USD/trades', function(req, res) {
  bitcoinUSDollarTrades(req.query, function(err, response) {
    if(err) {
      winston.error(err, " - /api/BTC/USD/trades", "(Spoof empty trade set) 200");
      res.send(200, []);
      return;
    }

    res.send(200, response);
  });
});

// Set required ETag for orderbook requests to a constant value.
app.head('/api/BTC/USD/orderbook', function(req, res) {
  res.set('ETag', 'not-supported');
  res.send(200);
});

// Returns an empty order book, since orderbooks require an issuer.
app.get('/api/BTC/USD/orderbook', function(req, res) {
  res.set('ETag', 'not-supported');
  res.send(200, {bids: [], asks: []});
});

//function to handle all incoming requests
function requestHandler(req, res) {
  var path = req.path.slice(5),
    time   = Date.now(),
    ip     = req.connection.remoteAddress,
    apiRoute, nSockets;
  
  if (path.indexOf('/') > 0) path = path.slice(0, path.indexOf('/'));
  
  winston.info(ip, "POST", path, "["+(new Date())+"]");
  apiRoute = path.replace(/_/g, "").toLowerCase();
  
  if (apiRoutes[apiRoute]) {
    
    nSockets = countSockets();
    monitor.logRequest(apiRoute, nSockets);
    
    if (nSockets >= maxSockets) return res.send(503, { error: "Service Unavailable"});
    
    apiRoutes[apiRoute](req.body, function(err, response){
      
      
      if (err) {
        winston.error(err, " - "+path, "(Server Error) 500");
        res.send(500, { error: err });
        return;
      }
      
      res.send(200, response); 
      time = (Date.now()-time)/1000;
      winston.info(ip, path, 200, "["+(new Date())+"]", time+"s");
      monitor.logResponseTime(time, apiRoute);
    });
   
  } else {
    
    winston.info("Response 404 Not Found - ", path);
    res.send(404, 'Sorry, that API route doesn\'t seem to exist.'+
      ' Available paths are: ' + 
      Object.keys(apiRoutes).join(', ') + '\n');
  }

  res.setTimeout(45 * 1000); //max 45s
  res.on("timeout", function(){
    winston.error("Response 408 Request Timeout - ", path);
    res.send(408, {error: "Request Timeout"});
  }); 
}

//initialize ledger monitor
monitor.ledgerMonitor();

//cache initialization
if (CACHE) {
  if (!config.redis || !config.redis.port || !config.redis.host) {
    CACHE = false;
    winston.error("Redis port and host are required");
    
  } else {
    redis = require("redis").createClient(config.redis.port, config.redis.host, config.redis.options);

    // Allow configuring a redis db number for global use.
    if(config.redis.options.db) {
      redis.select(config.redis.options.db);
    }
    
    //reset cache if the arg is present
    if (process.argv.indexOf('reset-cache') !== -1) redis.flushdb();
/*    
    redis.del("TVS:XPS:live:86400", function(err, res){
    });
    
    redis.del("TM:XPS:live:86400", function(err, res){
    });
    
    redis.del("TNV:XPS:live", function(err, res){
    });    
*/  
    redis.on("error", function (err) {
      winston.error("Redis - " + err);
      CACHE = false; //turn it off if its not working
    }); 
    
    //initialize historical metrics and associated cron jobs
    require('./library/history').init(); 
  } 
}


function countSockets () {
  var count = 0;
  for (var key1 in http.globalAgent.sockets) {
    count += http.globalAgent.sockets[key1].length;
  }
    
  for (var key2 in https.globalAgent.sockets) {
    count += https.globalAgent.sockets[key2].length;
  } 
  
  if (DEBUG) winston.info("open sockets: ", count);
  return count; 
} 